/*
 *  Universit degli Studi di Trento - Facolt di scienze
 *
 *  Author:  Paolo Larcheri
 *  Date:    May 2005
 *  
 */
/***********************************************************************
Class VBR_Traffic implements a Variable Bit Rate traffic source.
This traffic generator consists on a user-configurable number of states.
For each state the user must specify:
 - the transition probability towards other states
 - packet size (constant or chosen by an exponential ditribution)
 - interarrival packet time (either two values with respective probabilities
   or equally distributed whithin a range or chosen by an exponential
   truncated distribution with a limit of 1460 bytes)

Tcl example:
 set VBR [new Application/Traffic/VBR]
 $VBR num-states 3                     ;# this command must always precede the following ones
 # state 0
 $VBR state-interval 0 constant 0.001             ;# note: state numeration starts from 0
 $VBR state-packetsize 0 eq-distributed 1200 1460 ;# note: 1460 is the maximum allowed
 # state 1
 $VBR state-interval 1 exponential 0.01               ;# 0.01 is the mean value
 $VBR state-packetsize 1 two-values 1024 0.3 1240 0.7 ;# the sum of probabilities must be 1
 # state 2
 $VBR state-interval 2 constant 0.015
 $VBR state-packetsize 2 exponential 1200   ;# truncated at 1460 (1200 -> mean size)
 # transition probabilities
 $VBR trans-probability 0 1 0.5             ;# 0 is the starting state, 1 the target state...
 $VBR trans-probability 1 0 0.5             ;# so [0 1] is totally different from [1 0]
 # this command must always end all VBR settings
 # (this command accomplishes the evaluation of the 
 # self-transition probability of all states)
 $VBR end-config

 # Returns the mean data rate
 $VBR get-MeanDataRate

 # Returns the equivalent bitrate data rate, a rate between MeanDataRate and PeakDataRate
 # depending on par2 (see below)
 # par1 -> is the time sample used by algorithm to compute the Equiv Bitrate
 #         if par1 <= 0 then a time sample value depending on the number
 #                      of states and on the interarrival times is chosen
 # par2 -> is the probability that Generated DataRate might be greater than the 
 #         computed Equiv BitRate
 #         if par2 == 1 then the MeanDataRate is returned
 #         if par2 == 0 then the PeakDataRate is returned
 $VBR get-EquivDataRate -1 0.01

 # Returns the peak data rate
 $VBR get-PeakDataRate
***********************************************************************/
// HCF v3.0

#include <stdlib.h>
#include <string.h>

#include "vbr_state.h"
#include "trafgen.h"

class DFS_entry; // defined below

class VBR_Traffic : public TrafficGenerator {
 public:
   VBR_Traffic();
   virtual ~VBR_Traffic();
   virtual double next_interval(int&);
   //HACK so that udp agent knows interpacket arrival time within a burst
   inline double interval() { return (interval_); }

 protected:
   virtual int command(int argc, const char*const* argv);
   virtual void start();
   void init();
   void allocate_states(int n);
   void delete_states();
   int computeMeanDataRate();
   int computeEquivDataRate(double time_sample, double prob_bound);
   int computePeakDataRate();

   // Write in 'result' the stationary probabilities
   void computeStationaryProb(double* result);
   // Used to perform a DFS in the Markov Chain in order to
   // compute the Equivalent DataRate. 
   void DFS_EquivDataRate(double targ_time, double elaps_time, double prob,
                          VBR_state* state, int bits, list<DFS_entry> &result);

   double interval_; /* packet inter-arrival time during burst (sec) */
   int seqno_;
   int maxpkts_;

   int n_states;
   VBR_state **states;
   VBR_state *curr_state;
};


// Instances of this class are stored in a list while computing
// the Equivalent Data Rate. Each entry represents a feasible "navigation"
// of the Markov Chain having a given duration.
class DFS_entry {
public:
   double prob;
   int bitrate;  // bits/sec

   DFS_entry(double prob_, int bitrate_) : prob(prob_), bitrate(bitrate_)
   {}

   // This methd is needed since the list
   // of entries must be sorted.
   // Note: the ordering is performed considering the
   // DataRate of the navigation
   bool operator<(DFS_entry const &other)
   {
      if (this->bitrate < other.bitrate)
         return true;
      return false;
   }
};
